@@ -3,12 +3,13 @@ import socketio
 import logging
 import time
 from datetime import datetime
-import math
 import random
 from inspect import signature
 from typing import Union, Literal, Callable, List, Dict, Optional, TypeVar
 
 
+Any = object()
+
 DEVICE = 'device'
 DEVICES = 'devices'
 ALL_DATA = 'all_data'
@@ -48,6 +49,8 @@ EVENTS = Union[
     SET_NEW_DEVICE_NR
 ]
 
+INPUT_TYPE = Literal['text', 'number', 'datetime', 'date', 'time', 'select']
+
 
 def time_s() -> float:
     '''
@@ -93,7 +96,7 @@ class DictX(dict):
 
 
 class TimeStampedMsg(DictX):
-    time_stamp: int
+    time_stamp: float
 
 
 class BaseMsg(TimeStampedMsg):
@@ -112,12 +115,12 @@ class DevicesMsg(TimeStampedMsg):
     devices: List[Device]
 
 
-class DeviceJoinedMsg(DictX):
+class DeviceJoinedMsg(BaseMsg):
     room: str
     device: Device
 
 
-class DeviceLeftMsg(DictX):
+class DeviceLeftMsg(BaseMsg):
     room: str
     device: Device
 
@@ -125,7 +128,7 @@ class DeviceLeftMsg(DictX):
 class BaseSendMsg(DictX):
     device_id: Optional[str]
     device_nr: Optional[int]
-    time_stamp: Optional[int]
+    time_stamp: Optional[float]
 
 
 class DataMsg(BaseMsg):
@@ -134,10 +137,10 @@ class DataMsg(BaseMsg):
 
 class KeyMsg(DataMsg):
     type: Literal['key']
-    key: Literal['up', 'right', 'down', 'left', 'home']
+    key: Literal['up', 'right', 'down', 'left', 'home', 'f1', 'f2', 'f3', 'f4']
 
 
-class PointerData(DictX):
+class PointerData(BaseMsg):
     context: Literal['color', 'grid']
 
 
@@ -148,6 +151,7 @@ class ColorPointer(PointerData):
     y: int
     width: int
     height: int
+    displayed_at: float
 
 
 class GridPointer(PointerData):
@@ -156,16 +160,17 @@ class GridPointer(PointerData):
     row: int
     column: int
     color: str
+    displayed_at: float
 
 
-class Acc(DictX):
+class Acc(BaseMsg):
     x: float
     y: float
     z: float
     interval: float
 
 
-class Gyro(DictX):
+class Gyro(BaseMsg):
     alpha: float
     beta: float
     gamma: float
@@ -180,7 +185,7 @@ class GyroMsg(Gyro):
     type: Literal['gyro']
 
 
-class ErrorMsg(DictX):
+class ErrorMsg(BaseMsg):
     type: EVENTS
     msg: str
     err: Union[str, dict]
@@ -194,10 +199,12 @@ class InformationMsg(TimeStampedMsg):
 class InputResponseMsg(DataMsg):
     type: Literal['input_response']
     response: str
+    displayed_at: float
 
 
 class AlertConfirmMsg(DataMsg):
     type: Literal['alert_confirm']
+    displayed_at: float
 
 
 def flatten(list_of_lists: List[List]) -> List:
@@ -251,9 +258,10 @@ def first(filter_func: Callable[[T], bool], list_: List[T]) -> T:
 
 
 class Connector:
-    __start_time_ns: int = time_s()
+    __last_time_stamp: float = -1
+    __last_sub_time: float = 0
     data: Dict[str, List[BaseMsg]] = DictX({})
-    __devices: DevicesMsg = {'time_stamp': time.time(), 'devices': []}
+    __devices: DevicesMsg = {'time_stamp': time_s(), 'devices': []}
     device: Device = DictX({})
     __server_url: str
     __device_id: str
@@ -263,21 +271,22 @@ class Connector:
     joined_rooms: List[str]
 
     # callback functions
-    on_key: Callable[[KeyMsg, Optional[Connector]], None] = None
-    on_pointer: Callable[[Union[ColorPointer, GridPointer], Optional[Connector]], None] = None
-    on_acceleration: Callable[[AccMsg, Optional[Connector]], None] = None
-    on_gyro: Callable[[GyroMsg, Optional[Connector]], None] = None
-    on_sensor: Callable[[Union[GyroMsg, AccMsg], Optional[Connector]], None] = None
 
-    on_data: Callable[[DataMsg, Optional[Connector]], None] = None
-    on_broadcast_data: Callable[[DataMsg, Optional[Connector]], None] = None
-    on_all_data: Callable[[List[DataMsg], Optional[Connector]], None] = None
-    on_device: Callable[[Device, Optional[Connector]], None] = None
-    on_client_device: Callable[[Union[Device, None], Optional[Connector]], None] = None
-    on_devices: Callable[[List[Device], Optional[Connector]], None] = None
-    on_error: Callable[[ErrorMsg, Optional[Connector]], None] = None
-    on_room_joined: Callable[[DeviceJoinedMsg, Optional[Connector]], None] = None
-    on_room_left: Callable[[DeviceLeftMsg, Optional[Connector]], None] = None
+    on_key: Callable[[Any, KeyMsg, Optional[Connector]], None] = None
+    on_pointer: Callable[[Any, Union[ColorPointer, GridPointer], Optional[Connector]], None] = None
+    on_acceleration: Callable[[Any, AccMsg, Optional[Connector]], None] = None
+    on_gyro: Callable[[Any, GyroMsg, Optional[Connector]], None] = None
+    on_sensor: Callable[[Any, Union[GyroMsg, AccMsg], Optional[Connector]], None] = None
+
+    on_data: Callable[[Any, DataMsg, Optional[Connector]], None] = None
+    on_broadcast_data: Callable[[Any, DataMsg, Optional[Connector]], None] = None
+    on_all_data: Callable[[Any, List[DataMsg], Optional[Connector]], None] = None
+    on_device: Callable[[Any, Device, Optional[Connector]], None] = None
+    on_client_device: Callable[[Any, Union[Device, None], Optional[Connector]], None] = None
+    on_devices: Callable[[Any, List[Device], Optional[Connector]], None] = None
+    on_error: Callable[[Any, ErrorMsg, Optional[Connector]], None] = None
+    on_room_joined: Callable[[Any, DeviceJoinedMsg, Optional[Connector]], None] = None
+    on_room_left: Callable[[Any, DeviceLeftMsg, Optional[Connector]], None] = None
 
     __responses: List[InputResponseMsg] = []
     __alerts: List[AlertConfirmMsg] = []
@@ -294,6 +303,16 @@ class Connector:
     def device_id(self):
         return self.__device_id
 
+    @property
+    def current_time_stamp(self):
+        ts = time_s()
+        if ts == self.__last_time_stamp:
+            self.__last_sub_time += 0.000001
+            return ts + self.__last_sub_time
+        self.__last_sub_time = 0
+        self.__last_time_stamp = ts
+        return ts
+
     def __init__(self, server_url: str, device_id: str):
         self.__server_url = server_url
         self.__device_id = device_id
@@ -333,7 +352,7 @@ class Connector:
             the device number to which this message is sent exclusively. When set, boradcast has no effect. 
         '''
         if 'time_stamp' not in data:
-            data['time_stamp'] = time_s()
+            data['time_stamp'] = self.current_time_stamp
 
         if 'device_id' not in data:
             data['device_id'] = self.device_id
@@ -419,7 +438,7 @@ class Connector:
         unicast_to : int
             the device number to which this message is sent exclusively. When set, boradcast has no effect.
         '''
-        ts = time_s()
+        ts = self.current_time_stamp
 
         self.emit(
             ADD_NEW_DATA,
@@ -441,55 +460,77 @@ class Connector:
             alert_msg = next((res for res in self.__alerts if res['time_stamp'] == ts), False)
         self.__alerts.remove(alert_msg)
 
-    def input(self, question: str, input_type: Literal['text', 'number', 'datetime', 'date', 'time'] = 'text', unicast_to: int = None) -> Union[str, None]:
+    def input(self, question: str, input_type: INPUT_TYPE = 'text', options: List[str] = None, unicast_to: int = None) -> Union[str, None]:
         '''
         Parameters
         ----------
         question : str
             what should the user be prompted for?
 
-        input_type : 'text', 'number', 'datetime', 'date', 'time'
+        input_type : 'text', 'number', 'datetime', 'date', 'time', 'select'
             to use the correct html input type
 
         Optional
         --------
+        options: List[str]
+            required when input_type is 'select' - a list with the selection-options
         unicast_to : int
             the device number to which this message is sent exclusively.
 
         Return
         ------
-        str, int, float, datetime, date, dtime, None
-
-            depending on what you specify at return_type. When the response can not be casted, the raw value will be returned.
+        str, None
 
             When the user canceled the prompt, None is returned
         '''
-        return self.prompt(question, input_type=input_type, unicast_to=unicast_to)
+        return self.prompt(question, input_type=input_type, options=options, unicast_to=unicast_to)
 
-    def prompt(self, question: str, input_type: Literal['text', 'number', 'datetime', 'date', 'time'] = 'text', unicast_to: int = None) -> Union[str, None]:
+    def select(self, question: str, options: List[str]):
         '''
         Parameters
         ----------
         question : str
             what should the user be prompted for?
-
-        input_type : 'text', 'number', 'datetime', 'date', 'time'
-            to use the correct html input type
+        options: List[str]
+            a list with the options a user can select
 
         Optional
         --------
         unicast_to : int
             the device number to which this message is sent exclusively.
 
+        Retrun
+        ------
+        str, None
+
+            the selected value. None is returned when the prompt is canceled
+        '''
+        return self.prompt(question, input_type='select', options=options)
+
+    def prompt(self, question: str, input_type: INPUT_TYPE = 'text', options: List[str] = None, unicast_to: int = None) -> Union[str, None]:
+        '''
+        Parameters
+        ----------
+        question : str
+            what should the user be prompted for?
+
+        input_type : 'text', 'number', 'datetime', 'date', 'time', 'select'
+            to use the correct html input type
+
+        Optional
+        --------
+        options: List[str]
+            required when input_type is 'select' - a list with the selection-options
+        unicast_to : int
+            the device number to which this message is sent exclusively.
+
         Return
         ------
-        str, int, float, datetime, date, dtime, None
-
-            depending on what you specify at return_type. When the response can not be casted, the raw value will be returned.
+        str, None
 
             When the user canceled the prompt, None is returned 
         '''
-        ts = time_s()
+        ts = self.current_time_stamp
 
         if input_type == 'datetime':
             input_type = 'datetime-local'
@@ -500,6 +541,7 @@ class Connector:
                 'type': INPUT_PROMPT,
                 'question': question,
                 'input_type': input_type,
+                'options': options,
                 'time_stamp': ts
             },
             unicast_to=unicast_to
@@ -804,7 +846,7 @@ class Connector:
         ------
         bool wheter the assignment was succesfull or not.
         '''
-        ts = time.time()
+        ts = time_s()
         self.__info_messages.clear()
         self.emit(
             SET_NEW_DEVICE_NR,
@@ -816,7 +858,7 @@ class Connector:
             }
         )
         result_msg = None
-        while result_msg is None and (time.time() - ts) < max_wait:
+        while result_msg is None and (time_s() - ts) < max_wait:
             info_cnt = len(self.__info_messages)
 
             if info_cnt > 0 and self.__info_messages[info_cnt - 1].action['time_stamp'] == ts:
@@ -827,7 +869,7 @@ class Connector:
         if result_msg is not None and result_msg['message'] == 'Success':
             return True
 
-        time_left = max_wait - (time.time() - ts)
+        time_left = max_wait - (time_s() - ts)
         if time_left > 0 and 'should_retry' in result_msg and result_msg['should_retry']:
             return self.set_device_nr(new_device_nr, device_id=device_id, current_device_nr=current_device_nr, max_wait=time_left)
 
@@ -985,68 +1027,84 @@ class Connector:
 
 if __name__ == '__main__':
     logging.basicConfig(level=logging.DEBUG)
-    # connector = Connector('http://localhost:5000', 'FooBar')
-    connector = Connector('https://io.lebalz.ch', 'FooBar')
-    t0 = time.time()
-    # print('set deivce nr: ', connector.set_device_nr(13))
+    smartphone = Connector('http://localhost:5000', 'FooBar')
+    # smartphone = Connector('https://io.lebalz.ch', 'FooBar')
+    t0 = time_s()
+    smartphone.print(
+        """Hiiii
+        mother fucker
+        """)
+    smartphone.print('aasd1')
+    smartphone.print('aasd2')
+    smartphone.print('aasd3')
+    smartphone.print('aasd4')
+    smartphone.print('aasd5')
+    smartphone.print('aasd6')
+    smartphone.print('aasd7')
+    smartphone.print('aasd8')
+    smartphone.print('aasd9')
+    smartphone.print('aasd10')
+    response = smartphone.input("Hallo", input_type="select", options=["+", ":", "-", "*"])
+
+    # print('set deivce nr: ', smartphone.set_device_nr(13))
 
     # draw a 3x3 checker board
-    connector.set_grid([
+    smartphone.set_grid([
         ['black', 'white', 'black'],
         ['white', 'black', 'white'],
         ['black', 'white', 'black'],
     ], broadcast=True)
 
-    connector.on_key = lambda data, c: logging.info(f'on_key: {data}, len: {len(c.all_data())}')
-    connector.on_broadcast_data = lambda data: logging.info(f'on_broadcast_data: {data}')
-    connector.on_data = lambda data: logging.info(f'on_data: {data}')
-    connector.on_all_data = lambda data: logging.info(f'on_all_data: {data}')
-    connector.on_device = lambda data: logging.info(f'on_device: {data}')
-    connector.on_devices = lambda data: logging.info(f'on_devices: {data}')
-    connector.on_acceleration = lambda data: logging.info(f'on_acceleration: {data}')
-    connector.on_gyro = lambda data: logging.info(f'on_gyro: {data}')
-    connector.on_sensor = lambda data: logging.info(f'on_sensor: {data}')
-    connector.on_room_joined = lambda data: logging.info(f'on_room_joined: {data}')
-    connector.on_room_left = lambda data: logging.info(f'on_room_left: {data}')
-    connector.on_pointer = lambda data: logging.info(f'on_pointer: {data}')
-    connector.on_client_device = lambda data: logging.info(f'on_client_device: {data}')
-    connector.on_error = lambda data: logging.info(f'on_error: {data}')
+    smartphone.on_key = lambda data, c: logging.info(f'on_key: {data}, len: {len(c.all_data())}')
+    smartphone.on_broadcast_data = lambda data: logging.info(f'on_broadcast_data: {data}')
+    smartphone.on_data = lambda data: logging.info(f'on_data: {data}')
+    smartphone.on_all_data = lambda data: logging.info(f'on_all_data: {data}')
+    smartphone.on_device = lambda data: logging.info(f'on_device: {data}')
+    smartphone.on_devices = lambda data: logging.info(f'on_devices: {data}')
+    smartphone.on_acceleration = lambda data: logging.info(f'on_acceleration: {data}')
+    smartphone.on_gyro = lambda data: logging.info(f'on_gyro: {data}')
+    smartphone.on_sensor = lambda data: logging.info(f'on_sensor: {data}')
+    smartphone.on_room_joined = lambda data: logging.info(f'on_room_joined: {data}')
+    smartphone.on_room_left = lambda data: logging.info(f'on_room_left: {data}')
+    smartphone.on_pointer = lambda data: logging.info(f'on_pointer: {data}')
+    smartphone.on_client_device = lambda data: logging.info(f'on_client_device: {data}')
+    smartphone.on_error = lambda data: logging.info(f'on_error: {data}')
 
     time.sleep(2)
-    response = connector.input('Name? ')
-    connector.print(f'Name: {response} ')
-    connector.notify('notify hiii', alert=True)
-    print(connector.joined_room_count)
-    print(connector.client_count)
-    print(connector.device_count)
+    response = smartphone.input('Name? ')
+    smartphone.print(f'Name: {response} ')
+    smartphone.notify('notify hiii', alert=True)
+    print(smartphone.joined_room_count)
+    print(smartphone.client_count)
+    print(smartphone.device_count)
 
     print('\n')
-    print('data: ', connector.all_data())
-    print('data: ', connector.all_data(data_type='grid'))
-    print('latest data: ', connector.latest_data())
-    print('time_stamp', to_datetime(connector.latest_data()))
-    print('latest data: ', connector.latest_data(data_type='key'))
-    print('broadcast data: ', connector.all_broadcast_data())
-    print('broadcast data: ', connector.all_broadcast_data(data_type='grid'))
-    print('latest broadcast data: ', connector.latest_broadcast_data())
-    print('latest broadcast data: ', connector.latest_broadcast_data(data_type='key'))
-    print('cnt device', connector.device_count)
-    print('cnt room', connector.room_member_count)
-    print('cnt clients', connector.client_count)
-    print('cnt joined rooms', connector.joined_room_count)
-    print('pointer_data', connector.pointer_data())
-    print('data_list', connector.data_list)
-    print('color_pointer_data', connector.color_pointer_data())
-    print('grid_pointer_data', connector.grid_pointer_data())
-    print('gyro_data', connector.gyro_data())
-    print('acceleration_data', connector.acceleration_data())
-    print('key_data', connector.key_data())
-    print('latest_pointer', connector.latest_pointer())
-    print('latest_color_pointer', connector.latest_color_pointer())
-    print('latest_grid_pointer', connector.latest_grid_pointer())
-    print('latest_gyro', connector.latest_gyro())
-    print('latest_acceleration', connector.latest_acceleration())
-    print('latest_key', connector.latest_key())
+    print('data: ', smartphone.all_data())
+    print('data: ', smartphone.all_data(data_type='grid'))
+    print('latest data: ', smartphone.latest_data())
+    print('time_stamp', to_datetime(smartphone.latest_data()))
+    print('latest data: ', smartphone.latest_data(data_type='key'))
+    print('broadcast data: ', smartphone.all_broadcast_data())
+    print('broadcast data: ', smartphone.all_broadcast_data(data_type='grid'))
+    print('latest broadcast data: ', smartphone.latest_broadcast_data())
+    print('latest broadcast data: ', smartphone.latest_broadcast_data(data_type='key'))
+    print('cnt device', smartphone.device_count)
+    print('cnt room', smartphone.room_member_count)
+    print('cnt clients', smartphone.client_count)
+    print('cnt joined rooms', smartphone.joined_room_count)
+    print('pointer_data', smartphone.pointer_data())
+    print('data_list', smartphone.data_list)
+    print('color_pointer_data', smartphone.color_pointer_data())
+    print('grid_pointer_data', smartphone.grid_pointer_data())
+    print('gyro_data', smartphone.gyro_data())
+    print('acceleration_data', smartphone.acceleration_data())
+    print('key_data', smartphone.key_data())
+    print('latest_pointer', smartphone.latest_pointer())
+    print('latest_color_pointer', smartphone.latest_color_pointer())
+    print('latest_grid_pointer', smartphone.latest_grid_pointer())
+    print('latest_gyro', smartphone.latest_gyro())
+    print('latest_acceleration', smartphone.latest_acceleration())
+    print('latest_key', smartphone.latest_key())
 
-    connector.sleep(2)
-    connector.disconnect()
+    smartphone.sleep(2)
+    smartphone.disconnect()
